# AUTO-GENERATED FILE.
# Re-run /chalk/feature_n/codegen.py to regenerate contents.
from .feature_1.feature import Features as Feature1
from .feature_2.feature import Features as Feature2
from .feature_3.feature import Features as Feature3
from .feature_4.feature import Features as Feature4
from .feature_5.feature import Features as Feature5
from .feature_6.feature import Features as Feature6
from .feature_7.feature import Features as Feature7
from .feature_8.feature import Features as Feature8
from .feature_9.feature import Features as Feature9
from .feature_10.feature import Features as Feature10
from .feature_11.feature import Features as Feature11
from .feature_12.feature import Features as Feature12
from .feature_13.feature import Features as Feature13
from .feature_14.feature import Features as Feature14
from .feature_15.feature import Features as Feature15
from .feature_16.feature import Features as Feature16
from .feature_17.feature import Features as Feature17
from .feature_18.feature import Features as Feature18
from .feature_19.feature import Features as Feature19
from .feature_20.feature import Features as Feature20
from .feature_21.feature import Features as Feature21
from .feature_22.feature import Features as Feature22
from .feature_23.feature import Features as Feature23
from .feature_24.feature import Features as Feature24
from .feature_25.feature import Features as Feature25
from .feature_26.feature import Features as Feature26
from .feature_27.feature import Features as Feature27
from .feature_28.feature import Features as Feature28
from .feature_29.feature import Features as Feature29
from .feature_30.feature import Features as Feature30
from .feature_31.feature import Features as Feature31
from .feature_32.feature import Features as Feature32
from .feature_33.feature import Features as Feature33
from .feature_34.feature import Features as Feature34
from .feature_35.feature import Features as Feature35
from .feature_36.feature import Features as Feature36
from .feature_37.feature import Features as Feature37
from .feature_38.feature import Features as Feature38
from .feature_39.feature import Features as Feature39
from .feature_40.feature import Features as Feature40
from .feature_41.feature import Features as Feature41
from .feature_42.feature import Features as Feature42
from .feature_43.feature import Features as Feature43
from .feature_44.feature import Features as Feature44
from .feature_45.feature import Features as Feature45
from .feature_46.feature import Features as Feature46
from .feature_47.feature import Features as Feature47
from .feature_48.feature import Features as Feature48
from .feature_49.feature import Features as Feature49
from .feature_50.feature import Features as Feature50
from .feature_51.feature import Features as Feature51
from .feature_52.feature import Features as Feature52
from .feature_53.feature import Features as Feature53
from .feature_54.feature import Features as Feature54
from .feature_55.feature import Features as Feature55
from .feature_56.feature import Features as Feature56
from .feature_57.feature import Features as Feature57
from .feature_58.feature import Features as Feature58
from .feature_59.feature import Features as Feature59
from .feature_60.feature import Features as Feature60
from .feature_61.feature import Features as Feature61
from .feature_62.feature import Features as Feature62
from .feature_63.feature import Features as Feature63
from .feature_64.feature import Features as Feature64
from .feature_65.feature import Features as Feature65
from .feature_66.feature import Features as Feature66
from .feature_67.feature import Features as Feature67
from .feature_68.feature import Features as Feature68
from .feature_69.feature import Features as Feature69
from .feature_70.feature import Features as Feature70
from .feature_71.feature import Features as Feature71
from .feature_72.feature import Features as Feature72
from .feature_73.feature import Features as Feature73
from .feature_74.feature import Features as Feature74
from .feature_75.feature import Features as Feature75
from .feature_76.feature import Features as Feature76
from .feature_77.feature import Features as Feature77
from .feature_78.feature import Features as Feature78
from .feature_79.feature import Features as Feature79
from .feature_80.feature import Features as Feature80
from .feature_81.feature import Features as Feature81
from .feature_82.feature import Features as Feature82
from .feature_83.feature import Features as Feature83
from .feature_84.feature import Features as Feature84
from .feature_85.feature import Features as Feature85
from .feature_86.feature import Features as Feature86
from .feature_87.feature import Features as Feature87
from .feature_88.feature import Features as Feature88
from .feature_89.feature import Features as Feature89
from .feature_90.feature import Features as Feature90
from .feature_91.feature import Features as Feature91
from .feature_92.feature import Features as Feature92
from .feature_93.feature import Features as Feature93
from .feature_94.feature import Features as Feature94
from .feature_95.feature import Features as Feature95
from .feature_96.feature import Features as Feature96
from .feature_97.feature import Features as Feature97
from .feature_98.feature import Features as Feature98
from .feature_99.feature import Features as Feature99
from .feature_100.feature import Features as Feature100
from .feature_101.feature import Features as Feature101
from .feature_102.feature import Features as Feature102
from .feature_103.feature import Features as Feature103
from .feature_104.feature import Features as Feature104
from .feature_105.feature import Features as Feature105
from .feature_106.feature import Features as Feature106
from .feature_107.feature import Features as Feature107
from .feature_108.feature import Features as Feature108
from .feature_109.feature import Features as Feature109
from .feature_110.feature import Features as Feature110
from .feature_111.feature import Features as Feature111
from .feature_112.feature import Features as Feature112
from .feature_113.feature import Features as Feature113
from .feature_114.feature import Features as Feature114
from .feature_115.feature import Features as Feature115
from .feature_116.feature import Features as Feature116
from .feature_117.feature import Features as Feature117
from .feature_118.feature import Features as Feature118
from .feature_119.feature import Features as Feature119
from .feature_120.feature import Features as Feature120
from .feature_121.feature import Features as Feature121
from .feature_122.feature import Features as Feature122
from .feature_123.feature import Features as Feature123
from .feature_124.feature import Features as Feature124
from .feature_125.feature import Features as Feature125
from .feature_126.feature import Features as Feature126
from .feature_127.feature import Features as Feature127
from .feature_128.feature import Features as Feature128
from .feature_129.feature import Features as Feature129
from .feature_130.feature import Features as Feature130
from .feature_131.feature import Features as Feature131
from .feature_132.feature import Features as Feature132
from .feature_133.feature import Features as Feature133
from .feature_134.feature import Features as Feature134
from .feature_135.feature import Features as Feature135
from .feature_136.feature import Features as Feature136
from .feature_137.feature import Features as Feature137
from .feature_138.feature import Features as Feature138
from .feature_139.feature import Features as Feature139
from .feature_140.feature import Features as Feature140
from .feature_141.feature import Features as Feature141
from .feature_142.feature import Features as Feature142
from .feature_143.feature import Features as Feature143
from .feature_144.feature import Features as Feature144
from .feature_145.feature import Features as Feature145
from .feature_146.feature import Features as Feature146
from .feature_147.feature import Features as Feature147
from .feature_148.feature import Features as Feature148
from .feature_149.feature import Features as Feature149
from .feature_150.feature import Features as Feature150
from .feature_151.feature import Features as Feature151
from .feature_152.feature import Features as Feature152
from .feature_153.feature import Features as Feature153
from .feature_154.feature import Features as Feature154
from .feature_155.feature import Features as Feature155
from .feature_156.feature import Features as Feature156
from .feature_157.feature import Features as Feature157
from .feature_158.feature import Features as Feature158
from .feature_159.feature import Features as Feature159
from .feature_160.feature import Features as Feature160
from .feature_161.feature import Features as Feature161
from .feature_162.feature import Features as Feature162
from .feature_163.feature import Features as Feature163
from .feature_164.feature import Features as Feature164
from .feature_165.feature import Features as Feature165
from .feature_166.feature import Features as Feature166
from .feature_167.feature import Features as Feature167
from .feature_168.feature import Features as Feature168
from .feature_169.feature import Features as Feature169
from .feature_170.feature import Features as Feature170
from .feature_171.feature import Features as Feature171
from .feature_172.feature import Features as Feature172
from .feature_173.feature import Features as Feature173
from .feature_174.feature import Features as Feature174
from .feature_175.feature import Features as Feature175
from .feature_176.feature import Features as Feature176
from .feature_177.feature import Features as Feature177
from .feature_178.feature import Features as Feature178
from .feature_179.feature import Features as Feature179
from .feature_180.feature import Features as Feature180
from .feature_181.feature import Features as Feature181
from .feature_182.feature import Features as Feature182
from .feature_183.feature import Features as Feature183
from .feature_184.feature import Features as Feature184
from .feature_185.feature import Features as Feature185
from .feature_186.feature import Features as Feature186
from .feature_187.feature import Features as Feature187
from .feature_188.feature import Features as Feature188
from .feature_189.feature import Features as Feature189
from .feature_190.feature import Features as Feature190
from .feature_191.feature import Features as Feature191
from .feature_192.feature import Features as Feature192
from .feature_193.feature import Features as Feature193
from .feature_194.feature import Features as Feature194
from .feature_195.feature import Features as Feature195
from .feature_196.feature import Features as Feature196
from .feature_197.feature import Features as Feature197
from .feature_198.feature import Features as Feature198
from .feature_199.feature import Features as Feature199
from .feature_200.feature import Features as Feature200
from .feature_201.feature import Features as Feature201
from .feature_202.feature import Features as Feature202
from .feature_203.feature import Features as Feature203
from .feature_204.feature import Features as Feature204
from .feature_205.feature import Features as Feature205
from .feature_206.feature import Features as Feature206
from .feature_207.feature import Features as Feature207
from .feature_208.feature import Features as Feature208
from .feature_209.feature import Features as Feature209
from .feature_210.feature import Features as Feature210
from .feature_211.feature import Features as Feature211
from .feature_212.feature import Features as Feature212
from .feature_213.feature import Features as Feature213
from .feature_214.feature import Features as Feature214
from .feature_215.feature import Features as Feature215
from .feature_216.feature import Features as Feature216
from .feature_217.feature import Features as Feature217
from .feature_218.feature import Features as Feature218
from .feature_219.feature import Features as Feature219
from .feature_220.feature import Features as Feature220
from .feature_221.feature import Features as Feature221
from .feature_222.feature import Features as Feature222
from .feature_223.feature import Features as Feature223
from .feature_224.feature import Features as Feature224
from .feature_225.feature import Features as Feature225
from .feature_226.feature import Features as Feature226
from .feature_227.feature import Features as Feature227
from .feature_228.feature import Features as Feature228
from .feature_229.feature import Features as Feature229
from .feature_230.feature import Features as Feature230
from .feature_231.feature import Features as Feature231
from .feature_232.feature import Features as Feature232
from .feature_233.feature import Features as Feature233
from .feature_234.feature import Features as Feature234
from .feature_235.feature import Features as Feature235
from .feature_236.feature import Features as Feature236
from .feature_237.feature import Features as Feature237
from .feature_238.feature import Features as Feature238
from .feature_239.feature import Features as Feature239
from .feature_240.feature import Features as Feature240
from .feature_241.feature import Features as Feature241
from .feature_242.feature import Features as Feature242
from .feature_243.feature import Features as Feature243
from .feature_244.feature import Features as Feature244
from .feature_245.feature import Features as Feature245
from .feature_246.feature import Features as Feature246
from .feature_247.feature import Features as Feature247
from .feature_248.feature import Features as Feature248
from .feature_249.feature import Features as Feature249
from .feature_250.feature import Features as Feature250
from .feature_251.feature import Features as Feature251
from .feature_252.feature import Features as Feature252
from .feature_253.feature import Features as Feature253
from .feature_254.feature import Features as Feature254
from .feature_255.feature import Features as Feature255
from .feature_256.feature import Features as Feature256



from __future__ import annotations

import warnings
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Mapping, Optional, Sequence, Union

from chalk.clogging import chalk_logger
from chalk.features import DataFrame, Feature, FeatureConverter, Features, FeatureWrapper, ensure_feature
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql.protocols import BaseSQLSourceProtocol
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.string import normalize_string_for_matching

if TYPE_CHECKING:
    from sqlalchemy.sql import Select
    from sqlalchemy.sql.elements import TextClause


class Finalizer(str, Enum):
    ONE_OR_NONE = "OneOrNone"
    ONE = "One"
    FIRST = "First"
    ALL = "All"


def _get_matching_root_fqn(normalized_col_name: str, expected_features: Sequence[Feature]) -> Optional[str]:
    candidates: List[str] = []
    for x in expected_features:
        root_fqn_normalized = normalize_string_for_matching(x.root_fqn.lower())
        without_root_ns = normalize_string_for_matching(".".join(x.root_fqn.split(".")[1:]))
        if normalized_col_name == root_fqn_normalized or normalized_col_name == without_root_ns:
            candidates.append(x.root_fqn)
    if len(candidates) == 0:
        return None
    if len(candidates) > 1:
        # We really shouldn't hit this, unless if features are case-sensitive and the user is querying
        # snowflake which is case-insensitive
        raise ValueError(
            (
                f"Column '{normalized_col_name}' was ambiguous which feature it referred to. "
                f"Possible candidates were: {candidates}"
            )
        )
    return candidates[0]


class FinalizedChalkQuery:
    """A query that cannot be further filtered."""

    def __init__(
        self,
        query: Union[Select, TextClause],
        params: Mapping[str, Any],
        finalizer: Finalizer,
        incremental_settings: Optional[IncrementalSettings],
        source: BaseSQLSourceProtocol,
        fields: Mapping[str, Feature],
    ) -> None:
        self._query = query
        self._params = dict(params)
        self._finalizer = finalizer
        self._incremental_settings = incremental_settings
        self._source = source
        self._fields = fields

    @property
    def incremental_settings(self):
        return self._incremental_settings

    @property
    def finalizer(self):
        return self._finalizer

    @property
    def query(self):
        return self._query

    @property
    def source(self):
        return self._source

    @property
    def fields(self):
        return self._fields

    @property
    def params(self):
        return self._params

    def execute(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Any] = None,
    ) -> Union[Features, DataFrame, None]:
        """Actually execute the query to a `DataFrame` or set of features.

        If the finalizer was ONE, ONE_OR_NONE, or FIRST, then a `Features` instance
        is returned. Otherwise, if the `finalizer` is ALL, then a `DataFrame` instance
        is returned.

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resovler in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If `None` (default), a new connection will be acquired
            from the underlying source for the query

        Returns
        -------
        DataFrame
            A `DataFrame`, if the `.finalizer` is ALL; otherwise, `Features` set.
            If the finalizer is ONE_OR_NONE or FIRST, then the result may be
            `None` if no row was found
        """
        res = self.execute_to_dataframe(expected_features, connection)

        if self._finalizer in (Finalizer.ONE_OR_NONE, Finalizer.FIRST) and len(res) == 0:
            return None
        if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
            return res.slice(0, 1).to_features()[0]
        return res

    def execute_to_dataframe(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Any] = None,
    ) -> DataFrame:
        """Actually execute the query, and return a DataFrame. Unlike :meth:`.execute`, this method will always keep the
        results as a DataFrame, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resovler in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query

        Returns
        -------
        DataFrame
            A `DataFrame`, even if the result contains 0 or 1 rows.
        """
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        if self.incremental_settings is not None:
            # FIXME: Move the incrementalization logic here, so then the `execute` and `execute_to_dataframe`
            # methods can take the hwm timestamp as a paramater, to allow for direct execution
            warnings.warn(
                (
                    "This query specified an incremental configuration, which has not been applied. "
                    "This is likely because the resolver is being executed directly. "
                    "The query will be attempted without any high-water-mark timestamp. "
                    "This will attempt to select all data, "
                    "or if the filters depend on the incremental timestamp, "
                    "will result in a query execution error. "
                )
            )

        def converter_getters(column_names: List[str]) -> Dict[str, FeatureConverter]:
            # First map the column names to determine the feature fqns
            col_name_mapping = self.get_col_name_mapping(column_names, expected_features)
            return {k: Feature.from_root_fqn(v).converter for (k, v) in col_name_mapping.items()}

        pa_table = self.source.execute_query(self, converter_getters, connection)
        col_name_mapping = self.get_col_name_mapping(pa_table.column_names, expected_features)
        pa_table = pa_table.select(col_name_mapping.keys())
        pa_table = pa_table.rename_columns([col_name_mapping[x] for x in pa_table.column_names])
        df = pl.from_arrow(pa_table)
        assert isinstance(df, pl.DataFrame)

        if self._finalizer == Finalizer.ONE:
            if len(df) != 1:
                raise ValueError(f"Expected exactly one row; got {len(df)} rows")

        if self._finalizer == Finalizer.ONE_OR_NONE:
            if len(df) > 1:
                raise ValueError(f"Expected zero or one rows; got {len(df)} rows")

        if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
            df = df.slice(0, 1)

        return DataFrame(df)

    def execute_to_dataframe_batches(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Any] = None,
    ) -> Iterator[DataFrame]:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        if self.incremental_settings is not None:
            # FIXME: Move the incrementalization logic here, so then the `execute` and `execute_to_dataframe`
            # methods can take the hwm timestamp as a paramater, to allow for direct execution
            warnings.warn(
                (
                    "This query specified an incremental configuration, which has not been applied. "
                    "This is likely because the resolver is being executed directly. "
                    "The query will be attempted without any high-water-mark timestamp. "
                    "This will attempt to select all data, "
                    "or if the filters depend on the incremental timestamp, "
                    "will result in a query execution error. "
                )
            )

        def converter_getters(column_names: List[str]) -> Dict[str, FeatureConverter]:
            # First map the column names to determine the feature fqns
            col_name_mapping = self.get_col_name_mapping(column_names, expected_features)
            return {k: Feature.from_root_fqn(v).converter for (k, v) in col_name_mapping.items()}

        pa_tables = self.source.execute_query_to_batches(self, converter_getters, connection)
        i = 0
        for pa_table in pa_tables:
            i += 1
            chalk_logger.info(f"Processing batch {i} from query with shape {pa_table.shape}.")
            col_name_mapping = self.get_col_name_mapping(pa_table.column_names, expected_features)
            pa_table = pa_table.select(col_name_mapping.keys())
            pa_table = pa_table.rename_columns([col_name_mapping[x] for x in pa_table.column_names])
            df = pl.from_arrow(pa_table)
            assert isinstance(df, pl.DataFrame)

            if self._finalizer == Finalizer.ONE:
                if len(df) != 1:
                    raise ValueError(f"Expected exactly one row; got {len(df)} rows")

            if self._finalizer == Finalizer.ONE_OR_NONE:
                if len(df) > 1:
                    raise ValueError(f"Expected zero or one rows; got {len(df)} rows")

            if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
                df = df.slice(0, 1)

            yield DataFrame(df)

    def get_col_name_mapping(
        self,
        result_columns: List[str],
        expected_features: Optional[Sequence[Union[str, Feature, FeatureWrapper]]],
    ):
        """Map the output columns to the expected feature names.

        Parameters
        ----------
        result_columns
            A list of the columns, in order, returned by the query.
        expected_features
            The expected features for the query, as provided by the resolver signature.
            If a column name that is not in `fields` corresponds to a column in `expected_features`,
            then it will be mapped automatically.
            If a feature in `expected_features` does not have a corresponding output column,
            an error is raised.

        Returns
        -------
        dict[str, str]
            A mapping from output column names to root names.
        """
        ans: Dict[str, str] = {}
        normalized_to_original_col = {normalize_string_for_matching(x): x for x in result_columns}
        for k, v in self._fields.items():
            original_col_name = normalized_to_original_col.get(normalize_string_for_matching(k))
            if original_col_name is None:
                raise ValueError(f"Column {k} was not returned by the query.")
            ans[original_col_name] = v.root_fqn
        if expected_features is not None:
            unexpected_fields = [x for x in self._fields.values() if x not in expected_features]
            if len(unexpected_fields) > 0:
                raise ValueError(
                    f"Fields {unexpected_fields} were in the field mapping but are not included in the resolver output signature."
                )
            expected_features = [ensure_feature(y) for y in expected_features]
            for normalized_col_name, original_col_name in normalized_to_original_col.items():
                if original_col_name not in ans:
                    maybe_matching_root_fqn = _get_matching_root_fqn(normalized_col_name, expected_features)
                    if maybe_matching_root_fqn is not None:
                        ans[original_col_name] = maybe_matching_root_fqn
        return ans



from __future__ import annotations

import dataclasses
import inspect
import json
import os
import re
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterable, List, Literal, Optional, Type, Union, cast

import yaml
from pydantic import BaseModel, ValidationError, validator

from chalk import OfflineResolver, OnlineResolver
from chalk.features import DataFrame, Feature, Features, FeatureSetBase
from chalk.features.resolver import ResolverArgErrorHandler, StreamResolver
from chalk.sql import TableIngestProtocol
from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
from chalk.sql._internal.integrations.cloudsql import CloudSQLSourceImpl
from chalk.sql._internal.integrations.mysql import MySQLSourceImpl
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
from chalk.sql._internal.integrations.snowflake import SnowflakeSourceImpl
from chalk.sql._internal.integrations.sqlite import SQLiteFileSourceImpl
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.streams import KafkaSource
from chalk.streams.base import StreamSource
from chalk.streams.types import StreamResolverSignature
from chalk.utils.duration import parse_chalk_duration
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.string import to_camel_case

if TYPE_CHECKING:
    from sqlglot.expressions import Expression

_SOURCES = {
    "snowflake": SnowflakeSourceImpl,
    "postgres": PostgreSQLSourceImpl,
    "postgresql": PostgreSQLSourceImpl,
    "mysql": MySQLSourceImpl,
    "bigquery": BigQuerySourceImpl,
    "cloudsql": CloudSQLSourceImpl,
    "redshift": RedshiftSourceImpl,
    "sqlite": SQLiteFileSourceImpl,
    "kafka": KafkaSource,
}

_DIALECTS_SQLGLOT = {
    SnowflakeSourceImpl: "snowflake",
    PostgreSQLSourceImpl: "postgres",
    MySQLSourceImpl: "mysql",
    BigQuerySourceImpl: "bigquery",
    RedshiftSourceImpl: "redshift",
    SQLiteFileSourceImpl: "sqlite",
}

_RESOLVER_TYPES = {
    "offline": OfflineResolver,
    "batch": OfflineResolver,
    "online": OnlineResolver,
    "realtime": OnlineResolver,
    "stream": StreamResolver,
    "streaming": StreamResolver,
}


class IncrementalSettings(BaseModel):
    incremental_column: Optional[str]
    lookback_period: Optional[str]
    mode: Literal["row", "group", "parameter"]

    @validator("lookback_period")
    @classmethod
    def validate_lookback_period(cls, value):
        try:
            parse_chalk_duration(value)
        except Exception as e:
            raise ValueError(f"Could not parse value '{value}' as timedelta, {e}")
        return value

    @validator("mode")
    @classmethod
    def validate_mode(cls, mode, values):
        if mode in ["row", "group"] and not values["incremental_column"]:
            raise ValueError(f"'incremental_column' must be set if mode is 'row' or 'group'.")
        return mode


class CommentDict(BaseModel):
    source: str
    resolves: str
    incremental: Optional[IncrementalSettings]
    tags: Optional[List[str]]
    environment: Optional[List[str]]
    count: Optional[Literal[1, "one"]]
    cron: Optional[str]
    machine_type: Optional[str]
    max_staleness: Optional[str]
    message: Optional[str]
    owner: Optional[str]
    type: Optional[str]
    timeout: Optional[str]

    @validator("tags", "environment", pre=True)
    @classmethod
    def validate_list_inputs(cls, value):
        if isinstance(value, list):
            return value
        if isinstance(value, str):
            return [value]
        raise ValueError(f"Value {value} must be a string or a list of strings.")

    @validator("cron", "max_staleness", "timeout")
    @classmethod
    def validate_timedelta(cls, string):
        try:
            parse_chalk_duration(string)
        except Exception as e:
            raise ValueError(f"Could not parse value '{string}' as timedelta, {e}")
        return string

    @validator("type")
    @classmethod
    def validate_type(cls, resolver_type):
        if resolver_type not in _RESOLVER_TYPES:
            raise ValueError(
                f"Resolver type '{resolver_type}' not supported. "
                f"'online', 'offline' and 'streaming' are supported options"
            )
        return resolver_type


@dataclasses.dataclass
class ResolverError:
    """Generic class for returning errors at any point during resolution process"""

    display: str
    path: str
    parameter: Optional[str]


@dataclasses.dataclass
class ResolverResult:
    """Chief return class with resolver we actually use"""

    resolver: Optional[OnlineResolver]
    errors: List[ResolverError]
    db: Optional[TableIngestProtocol]
    fields: Optional[Dict[str, str]]
    args: Optional[Dict[str, str]]


@dataclasses.dataclass
class SQLStringResult:
    """Class for getting the sql string from the file"""

    path: str
    sql_string: Optional[str]
    error: Optional[ResolverError]

    @classmethod
    def fail(cls, display_error: str, path: str) -> "SQLStringResult":
        return cls(path=path, sql_string=None, error=ResolverError(display=display_error, path=path, parameter=None))


@dataclasses.dataclass
class GlotResult:
    """Class for editing the sql string, and using sqlglot on sql string"""

    sql_string: str
    glot: Optional[Expression]
    args: Optional[Dict[str, str]]
    default_args: List[Optional[str]]
    comment_dict: Optional[CommentDict]
    docstring: Optional[str]
    errors: List[ResolverError]


@dataclasses.dataclass
class ParseResult:
    """Class for important info gathered from glot"""

    sql_string: str
    comment_dict: Optional[CommentDict]
    fields: Optional[Dict[str, str]]
    namespace: Optional[str]
    source: Optional[Union[BaseSQLSource, StreamSource]]
    docstring: Optional[str]
    errors: List[ResolverError]


def get_sql_file_resolvers(
    sql_file_resolve_location: Path, sources: Iterable[BaseSQLSource]
) -> Iterable[ResolverResult]:
    """Iterate through all `.chalk.sql` filepaths, gather the sql strings, and get a resolver hopefully for each."""
    for dp, dn, fn in os.walk(os.path.expanduser(sql_file_resolve_location)):
        for f in fn:
            filepath = os.path.join(dp, f)
            if not filepath.endswith(".chalk.sql"):
                continue
            sql_string_result = _get_sql_string(filepath)
            yield get_sql_file_resolver(sources, sql_string_result)


def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: SQLStringResult) -> ResolverResult:
    """Parse the sql strings and get a ResolverResult from each"""
    if sql_string_result.error:
        return ResolverResult(resolver=None, errors=[sql_string_result.error], db=None, fields=None, args=None)
    path = sql_string_result.path

    errors: List[ResolverError] = []
    glot_result: GlotResult = _get_sql_glot(sql_string=sql_string_result.sql_string, path=path, sources=sources)
    if glot_result.errors:
        return ResolverResult(resolver=None, errors=glot_result.errors, db=None, fields=None, args=None)

    parsed: ParseResult = _parse_glot(glot_result, path, sources)
    if parsed.errors:
        return ResolverResult(resolver=None, errors=parsed.errors, db=parsed.source, fields=None, args=None)

    # validate inputs and outputs as real features in graph
    inputs: List[Feature] = []
    for arg in glot_result.args.values():
        try:
            inputs.append(Feature.from_root_fqn(arg))
        except Exception as e:
            errors.append(
                ResolverError(
                    display=f"The file '{path}' references an input feature '{arg}' which does not exist",
                    path=path,
                    parameter=arg,
                )
            )
    outputs: List[Feature] = []
    for output in parsed.fields.values():
        if output.endswith("*"):
            features = FeatureSetBase.registry[parsed.namespace].features
            for feature in features:
                if feature.is_scalar:
                    outputs.append(feature)
            continue
        try:
            outputs.append(Feature.from_root_fqn(output))
        except Exception as e:
            errors.append(
                ResolverError(
                    display=f"The file '{path}' references an output feature '{output}' which does not exist",
                    path=path,
                    parameter=output,
                )
            )

    resolver_type_str = parsed.comment_dict.type if parsed.comment_dict.type else "online"
    resolver_type = _RESOLVER_TYPES[resolver_type_str]

    if resolver_type == StreamResolver:
        result: ResolverResult = _get_stream_resolver(path, glot_result, parsed, outputs)
        return result

    result = ResolverResult(resolver=None, errors=errors, db=parsed.source, fields=None, args=None)

    incremental_dict = parsed.comment_dict.incremental.dict() if parsed.comment_dict.incremental else None
    return_one = parsed.comment_dict.count

    # function for online resolver to process
    def fn(
        *input_values,
        database=parsed.source,
        sql_query=parsed.sql_string,
        field_dict=parsed.fields,
        args_dict=glot_result.args,
        incremental=incremental_dict,
    ):
        arg_dict = {arg: input_value for input_value, arg in zip(input_values, args_dict.keys())}
        func = database.query_string(
            query=sql_query,
            fields=field_dict,
            args=arg_dict,
        )
        if incremental:
            func = func.incremental(**incremental)
        if return_one:
            func = func.one()
        return func

    if result.errors:
        return result

    if return_one:
        output = Features[tuple(outputs)]
    else:
        output = Features[DataFrame[tuple(outputs)]]

    default_args = [ResolverArgErrorHandler(default_value) for default_value in glot_result.default_args]

    filename = os.path.basename(path)
    # attempt to instantiate the resolver
    try:
        assert resolver_type in (OnlineResolver, OfflineResolver)
        resolver = resolver_type(
            filename=path,
            function_definition=_remove_comments(sql_string_result.sql_string),
            fqn=filename.replace(".chalk.sql", ""),
            doc=parsed.docstring,
            inputs=inputs,
            output=output,
            fn=fn,
            environment=parsed.comment_dict.environment,
            tags=parsed.comment_dict.tags,
            max_staleness=parsed.comment_dict.max_staleness,
            cron=parsed.comment_dict.cron,
            machine_type=parsed.comment_dict.machine_type,
            when=None,
            state=None,
            default_args=default_args,
            owner=parsed.comment_dict.owner,
            timeout=parsed.comment_dict.timeout,
        )
    except Exception as e:
        result.errors.append(
            ResolverError(
                display=f"{resolver_type_str.capitalize()} resolver could not be instantiated, {e}",
                path=path,
                parameter=None,
            )
        )
        return result

    result.resolver = resolver
    result.fields = parsed.fields
    result.args = glot_result.args
    return result


def _get_sql_string(path: str) -> SQLStringResult:
    """Attempt to get a sql string from a filepath and gracefully exit if unable to"""
    sql_string_result = SQLStringResult(path=path, sql_string=None, error=None)
    if not path.endswith(".chalk.sql"):
        return SQLStringResult.fail(display_error=f"sql resolver file '{path}' must end in '.chalk.sql'", path=path)
    sql_string = None
    if os.path.isfile(path):
        with open(path) as f:
            sql_string = f.read()
    else:
        caller_filename = inspect.stack()[1].filename
        dir_path = os.path.dirname(os.path.realpath(caller_filename))
        if isinstance(path, bytes):
            path = path.decode("utf-8")
        relative_path = os.path.join(dir_path, path)
        if os.path.isfile(relative_path):
            with open(relative_path) as f:
                sql_string = f.read()
    if sql_string is None:
        return SQLStringResult.fail(display_error=f"Cannot find file '{path}'", path=path)
    sql_string_result.sql_string = sql_string
    return sql_string_result


def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource]) -> GlotResult:
    """Get sqlglot from sql string and gracefully exit if unable to"""
    try:
        import sqlglot
        import sqlglot.expressions
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    glot_result = GlotResult(
        sql_string=sql_string, glot=None, args=None, default_args=[], comment_dict=None, docstring=None, errors=[]
    )
    args = {}  # sql string -> input feature string
    variables = set(re.findall("\\${.*?\\}", sql_string))
    # replace ?{variable_name} with :variable_name for sqlalchemy, and keep track of input args necessary
    for variable_pattern in variables:
        has_default_arg = False
        variable = variable_pattern[2:-1]  # cut off ${ and }
        for split_var in ["|", " or "]:  # default argument
            # TODO cannot parse something like {Transaction.category or "Waffles or Pancakes"} yet
            if split_var in variable:
                split = variable.split(split_var)
                if len(split) != 2:
                    glot_result.errors.append(
                        ResolverError(
                            display=f"If character '|' is used, both variable name and default value must be "
                            f'specified in ({variable}) like \'?{{variable_name | "default_value"}}',
                            path=path,
                            parameter=None,
                        )
                    )
                else:  # has default argument
                    variable = split[0].strip()
                    default_arg = split[1].strip()
                    default_arg = json.loads(default_arg)
                    f = Feature.from_root_fqn(variable)
                    default_arg = f.converter.from_json_to_rich(default_arg)
                    glot_result.default_args.append(default_arg)
                    has_default_arg = True
        if not has_default_arg:
            # TODO default value of None should be refactored since None is a legitimate default value
            glot_result.default_args.append(None)
        period_replaced = variable.replace(".", "_")
        sql_safe_str = f"__chalk_{period_replaced}__"
        sql_string = sql_string.replace(variable_pattern, f":{sql_safe_str}")
        args[sql_safe_str] = variable

    glot_result.sql_string = sql_string
    glot_result.args = args

    comments = ""
    docstring = ""
    for comment in glot_result.sql_string.splitlines():
        if comment.strip().startswith("--"):
            comment = comment.strip().replace("--", "")
            if comment.strip().startswith("-"):
                comments += f"{comment}\n"
            else:
                split = comment.split(":")
                if len(split) != 2:
                    docstring += f"{comment.strip()}\n"
                else:
                    comments += f"{comment}\n"
        else:
            break

    try:
        comment_dict = yaml.safe_load(comments)
    except Exception as e:
        glot_result.errors.append(
            ResolverError(
                display=f"Comments key-values '{comments}' must be of YAML form, {e}",
                path=path,
                parameter=comments,
            )
        )
        return glot_result
    try:
        comment_dict_object = CommentDict.parse_obj(comment_dict)
    except ValidationError as e:
        for error in e.errors():
            glot_result.errors.append(
                ResolverError(
                    display=f"Could not parse comment(s) '{'.'.join(error['loc'])}': {error['msg']}",
                    path=path,
                    parameter=comment_dict,
                )
            )
        return glot_result
    glot_result.comment_dict = comment_dict_object
    glot_result.docstring = docstring.strip()

    source_comment = glot_result.comment_dict.source
    source_type = None
    if source_comment not in _SOURCES:  # actual name of source
        for possible_source in sources:
            if possible_source.name == source_comment:
                source_type = _DIALECTS_SQLGLOT.get(type(possible_source))  # dialect may be unsupported
    else:
        if source_comment in _DIALECTS_SQLGLOT.values():  # source type, if supported
            source_type = source_comment
    try:
        glot_result.glot = sqlglot.parse_one(sql=glot_result.sql_string, read=source_type)
    except Exception as e:
        glot_result.errors.append(
            ResolverError(display=f"Cannot SQL parse {glot_result.sql_string}, {e}", path=path, parameter=None)
        )
        return glot_result
    if not isinstance(glot_result.glot, sqlglot.expressions.Select):
        glot_result.errors.append(
            ResolverError(
                display=f"SQL query {glot_result.sql_string} should be of 'SELECT' type", path=path, parameter=None
            )
        )
    return glot_result


def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSource]) -> ParseResult:
    """Parse useful info from sqlglot and gracefully exit if unable to"""
    parse_result = ParseResult(
        sql_string=glot_result.sql_string,
        comment_dict=None,
        fields=None,
        namespace=None,
        source=None,
        docstring=None,
        errors=[],
    )

    # define a source SQL database. Can either specify name or kind if only one of the kind is present.
    parse_result.comment_dict = glot_result.comment_dict
    parse_result.docstring = glot_result.docstring
    source_name = parse_result.comment_dict.source
    source = None
    if source_name not in _SOURCES:  # actual name of source
        for possible_source in sources:
            if possible_source.name == source_name:
                source = possible_source
    else:
        for possible_source in sources:
            if isinstance(possible_source, _SOURCES.get(source_name)):
                if source:
                    parse_result.errors.append(
                        ResolverError(
                            display=f"More than one {source_name} source exists. Instead, refer to the integration by "
                            f"name among ({[source.name for source in sources]}).",
                            path=path,
                            parameter=source_name,
                        )
                    )
                source = possible_source
    if not source:
        parse_result.errors.append(
            ResolverError(display=f"Source '{source_name}' not found", path=path, parameter=source_name)
        )
    parse_result.source = source

    if parse_result.errors:
        return parse_result

    # get resolver fields: which columns selected will match to which chalk feature?
    namespace = parse_result.comment_dict.resolves
    if namespace not in FeatureSetBase.registry:
        camelcase_namespace = to_camel_case(namespace)
        if camelcase_namespace in FeatureSetBase.registry:
            namespace = camelcase_namespace
        else:
            parse_result.errors.append(
                ResolverError(
                    display=f"No @features class with the name '{namespace}'", path=path, parameter=source_name
                )
            )
    parse_result.namespace = namespace
    fields = {}  # sql string -> output feature string

    for column_name in glot_result.glot.named_selects:
        fields[column_name] = f"{namespace}.{column_name}"
    parse_result.fields = fields

    return parse_result


def _get_stream_resolver(
    path: str, glot_result: GlotResult, parsed: ParseResult, outputs: List[Feature]
) -> ResolverResult:
    errors = []
    result = ResolverResult(resolver=None, errors=errors, db=parsed.source, fields=None, args=None)

    output_features = Features[DataFrame[tuple(outputs)]]

    if isinstance(output_features.features[0], type) and issubclass(output_features.features[0], DataFrame):
        output_feature_fqns = set(f.fqn for f in cast(Type[DataFrame], output_features.features[0]).columns)
    else:
        output_feature_fqns = set(f.fqn for f in output_features.features)

    signature = StreamResolverSignature(
        params=[],
        output_feature_fqns=output_feature_fqns,
    )

    sql_query: str = _remove_comments(parsed.sql_string)
    filename = os.path.basename(path)

    try:

        def fn():
            return sql_query

        # attempt to instantiate the resolver
        resolver = StreamResolver(
            function_definition=sql_query,
            fqn=filename.replace(".chalk.sql", ""),
            filename=path,
            source=parsed.source,
            fn=fn,
            environment=parsed.comment_dict.environment,
            doc=parsed.docstring,
            module=filename.replace(".chalk.sql", ""),
            mode=None,
            machine_type=parsed.comment_dict.machine_type,
            message=parsed.comment_dict.message,
            output=output_features,
            signature=signature,
            state=None,
            sql_query=sql_query,
            owner=parsed.comment_dict.owner,
            parse=None,
        )
    except Exception as e:
        result.errors.append(
            ResolverError(
                display=f"Streaming resolver could not be instantiated, {e}",
                path=path,
                parameter=None,
            )
        )
        return result

    result.resolver = resolver
    result.fields = parsed.fields
    result.args = glot_result.args
    return result


def _remove_comments(sql_string: str) -> str:
    sql_string = re.sub(
        re.compile("/\\*.*?\\*/", re.DOTALL), "", sql_string
    )  # remove all occurrences streamed comments (/*COMMENT */) from string
    sql_string = re.sub(
        re.compile("//.*?\n"), "", sql_string
    )  # remove all occurrence single-line comments (//COMMENT\n ) from string
    sql_string = re.sub(
        re.compile("--.*?\n"), "", sql_string
    )  # remove all occurrence single-line comments (//COMMENT\n ) from string
    return sql_string.strip()



from typing import List, Tuple, Union

from typing_extensions import TypeAlias

Tags: TypeAlias = Union[List[str], Tuple[str, ...], str]
"""
Tags allow you to scope requests within an
environment. Both tags and environment need
to match for a resolver to be a candidate to
execute.

Like Environments, tags control when resolvers
run based on the Online Context or Training Context
matching the tags provided to the resolver decorator.
Resolvers optionally take a keyword argument named
tags that can take one of three types:
- `None` (default) - The resolver will be a candidate to run for every set of tags.
- `str` - The resolver will run only if this tag is provided.
- `list[str]` - The resolver will run in all of the specified tags match.

See more at https://docs.chalk.ai/docs/resolver-tags
"""

Environments: TypeAlias = Union[List[str], Tuple[str, ...], str]
"""
Environments are used to trigger behavior
in different deployments such as staging,
production, and local development.
For example, you may wish to interact with
a vendor via an API call in the production
environment, and opt to return a constant
value in a staging environment.

`Environments` can take one of three types:
  - `None` (default) - candidate to run in every environment
  - `str` - run only in this environment
  - `list[str]` - run in any of the specified environment and no others

See more at https://docs.chalk.ai/docs/resolver-environments
"""

EnvironmentId: TypeAlias = str
"""Many of the method on the `ChalkClient` expose an `environment`
parameter for specifying which Chalk environment to target, as defined in
https://docs.chalk.ai/docs/resolver-environments

If you pass an explicit variable for environments, that will always be preferred.
Otherwise, the `ChalkClient` will first check for the presence of the environment
variables `CHALK_CLIENT_ID` and `CHALK_CLIENT_SECRET`. If they are present,
`ChalkClient` will use those credentials, and optionally the API server
and environment specified by `CHALK_API_SERVER` and `CHALK_ENVIRONMENT`.

Failing these two checks, the `ChalkClient` will look for a `~/.chalk.yml` file,
which is updated when you run `chalk login`. If a token for the specific project
directory if found, that token will be used. Otherwise, the token under the key
`default` will be used. When using the `~/.chalk.yml` file, you can still
optionally override the API server and environment by setting the environment
variables `CHALK_API_SERVER` and `CHALK_ENVIRONMENT`.
"""

BranchId: TypeAlias = str



# AUTO-GENERATED FILE.
# Re-run /chalk/feature_n/codegen.py to regenerate contents.
from typing import TypeVar, Generic, Optional, Dict
from chalk.features.dataframe import DataFrameMeta


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")
T11 = TypeVar("T11")
T12 = TypeVar("T12")
T13 = TypeVar("T13")
T14 = TypeVar("T14")
T15 = TypeVar("T15")
T16 = TypeVar("T16")
T17 = TypeVar("T17")


class Features(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17]):
    pass


class DataFrame(
    Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17], metaclass=DataFrameMeta
):
    def __getitem__(self, item):
        pass
